//////////////////////////////////////////////////////////////////////
// Fir.cpp: implementation of the CFir class.
//////////////////////////////////////////////////////////////////////
//SoundCard/Wavefile signal generator program
// Copyright 1999.    Moe Wheatley AE4JY  <ae4jy@mindspring.com>
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Fir.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CFir::CFir()
{
	m_pQue = NULL;
}

CFir::~CFir()
{
	if(m_pQue)
		delete m_pQue;
}

/////////////////////////////////////////////////////////////////
//            initialize FIR buffers and pointers
// n == FIR length
// pCoef == pointer to FIR coeficients
/////////////////////////////////////////////////////////////////
BOOL CFir::Init(INT n, const double* pCoef)
{
//circular delay line.(data stays put and the pointers move)
	if( (m_pQue = new double[sizeof(double)*n] ) == NULL)
		return FALSE;
	m_pCoef = pCoef;
	m_N = n;
	for(INT i=0; i<m_N; i++)
		m_pQue[i] = 0.0;		// fill delay buffer with zero
	m_pInPtr = m_pQue;
	return TRUE;
}

/////////////////////////////////////////////////////////////////
//			processes FIR filter on 'BlockSize' samples
//	pIn == pointer to input array of doubles (can be same buffer as pOut )
//	pOut == pointer to input array of doubles
//	Blocksize == number of samples to process
/////////////////////////////////////////////////////////////////
void CFir::ProcFIR( double* pIn, double* pOut , INT BlockSize)
{
INT i;
INT n;
double acc;
const double* Kptr;
double* Firptr;
double* Qptr;
double* Inptr;
	Inptr = m_pInPtr;		//use local copies of member variables
	Qptr = m_pQue;			// for better speed.
	n = m_N;
	for( i = 0; i<BlockSize; i++ )	// put new samples into Queue
	{
		if( --Inptr < Qptr )		//deal with wraparound
			Inptr = Qptr+n-1;
		*Inptr = pIn[i];
		acc = 0.0;
		Firptr = Inptr;
		Kptr = m_pCoef;
		while( Kptr < m_pCoef + n)	//do the MAC's
		{
			acc += ( (*Firptr++)*(*Kptr++) );
			if( Firptr >= Qptr+n )	//deal with wraparound
				Firptr = Qptr;
		}
		pOut[i] = acc;		//save output sample
	}
	m_pInPtr = Inptr;		// save position in circular delay line
}
