// MSKGen.cpp: implementation of the CMSKGen class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "wscgen.h"
#include "MSKGen.h"
#include "math.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define RMSCONSTANT 1.0
#define SYMBOL_RATE 31.25		// 31.25 Symbols per Second
#define FREQ_SHIFT (SYMBOL_RATE/4.0)	

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMSKGen::CMSKGen()
{
	m_2PI = 8.0 * atan(1.0);		//make constant 2 PI

}

CMSKGen::~CMSKGen()
{

}

/////////////////////////////////////////////////////////////
//Initialize the MSK generator with the sample frequncy Fs,
// the carrier frequency freq.
/////////////////////////////////////////////////////////////
void CMSKGen::InitMSK(INT Fs, double freq)
{
	SetMSKFreq( Fs, freq,0,0);	//set carrier frequency
	m_MSKSecPerSamp = 1.0/Fs;
	m_MSKTime = 0;
	m_MSKPeriodUpdate = 1.0/SYMBOL_RATE;

}

/////////////////////////////////////////////////////////////
// Call to change the psk31 carrier frequency
/////////////////////////////////////////////////////////////
void CMSKGen::SetMSKFreq(INT Fs, double freq, INT rate, INT range)
{
	m_MSKPhaseIncH = m_2PI * freq/(double)Fs;		//carrier frequency
	m_MSKPhaseIncL = m_2PI * freq/(double)Fs;		//carrier frequency
	m_MSKPhaseIncH += (m_2PI * FREQ_SHIFT/(double)Fs);
	m_MSKPhaseIncL -= (m_2PI * FREQ_SHIFT/(double)Fs);
}

/////////////////////////////////////////////////////////////
// generates n samples of MSK waveform in pData with average RMS value of 1.0
/////////////////////////////////////////////////////////////
BOOL CMSKGen::CalcMSK(double *pData, INT n)
{
static double tc = 0.0;
INT i;
INT mod;
	for( i=0; i<n; i++ )
	{
		m_MSKTime += m_MSKSecPerSamp;
		if( m_MSKTime >= m_MSKPeriodUpdate )
		{
			if(rand() >= 16384)
				mod = 1;
			else
				mod = -1;
			m_MSKTime -= m_MSKPeriodUpdate;	//keep time bounded
		}
		if(mod>0)
			tc = tc + (m_MSKPhaseIncH);	// increment radian phase count
		else
			tc = tc + (m_MSKPhaseIncL);	// increment radian phase count
		tc = fmod(tc,m_2PI);			//keep radian counter bounded
		pData[i] = RMSCONSTANT*( sin( tc ) + cos( tc ) );
	}
	return FALSE;
}