// ProcThread.cpp: implementation of the CProcThread class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ProcThread.h"
#include "math.h"


// Non Class routine for launching the worker thread

static UINT ThreadLauncher( CProcThread* pIO)
{
	pIO->ProcessLoop();
	return(4);
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CProcThread::CProcThread()
{
	m_pProcThread = NULL;
	m_fProcThreadQuit = FALSE;
	m_pSoundDev = NULL;
	m_pWaveDev = NULL;
	m_pGenSrc = NULL;
	m_pPSKSrc = NULL;
   m_pRTTYSrc = NULL;
	m_pSSTVSrc = NULL;
	m_p400BPSKSrc = NULL;
	m_pMSKSrc = NULL;
	m_pMFSKSrc = NULL;
	m_pOutBuf = NULL;
}

CProcThread::~CProcThread()
{

}

///////////////////////////////////////////////////////////////////////////
//  Stop process thread proceedure and wait for it to finish
//  Called by main program thread to signal data processing thread to stop
//	and then wait for it to stop before returning.
///////////////////////////////////////////////////////////////////////////
void CProcThread::StopThread()
{
	if( m_pProcThread != NULL )		//if thread is already running
	{
		m_fProcThreadQuit = TRUE;	// stop it
		::WaitForSingleObject(m_pProcThread->m_hThread,	// drawing thread handle
								INFINITE);
		delete m_pProcThread;
		m_pProcThread = NULL;
	}
}

///////////////////////////////////////////////////////////////////////////
//  Free resources and exit process thread
//   Called by process thread itself to shut itself off if not signaled by
//  main program thread.
///////////////////////////////////////////////////////////////////////////
void CProcThread::StopMyself()
{
//  if any resources are open, kill em.
	DeleteResources();
	m_pProcThread->m_bAutoDelete = TRUE;	//make thread clean up after
	m_pProcThread = NULL;  					// itself before it dies
	m_pDlg->ProcessAbort();					//let main thread know I'm dead
	AfxEndThread(3); 
}

///////////////////////////////////////////////////////////////////////////
//   Called to start the thread running
///////////////////////////////////////////////////////////////////////////
void CProcThread::StartThread(void* pDlg )
{
	m_pDlg = (CWSCGenDlg*)pDlg;				//save off pointer to parent class
	if( m_pProcThread == NULL)
	{
		m_fProcThreadQuit = FALSE;
		m_pProcThread = AfxBeginThread(
			(AFX_THREADPROC)ThreadLauncher,		//thread function
			(LPVOID)this,						// ptr to the class
			THREAD_PRIORITY_NORMAL, 			//give worker thread priority
			0,									// same stack size
			CREATE_SUSPENDED,					// don't let it start yet
			NULL );							// same security attributes
		m_pProcThread->m_bAutoDelete = FALSE;	// keep thread object around
		m_pProcThread->ResumeThread();	//let er rip
	}
}

///////////////////////////////////////////////////////////////////////////
//  Main I/O thread processing loop
///////////////////////////////////////////////////////////////////////////
void CProcThread::ProcessLoop()
{
WAVEFORMATEX Wfx;
INT BufSize;
DWORD dwOutBlockSize;
DWORD SampleLimit;
BOOL fFinished = FALSE;
BOOL result;
DWORD SamplesSaved;
UINT ErrorCode;

	Wfx.wFormatTag = WAVE_FORMAT_PCM;
	Wfx.nChannels = 1;					//don't use stereo
	Wfx.nSamplesPerSec = m_pDlg->m_SampsPerSec;
	Wfx.nAvgBytesPerSec = m_pDlg->m_SampsPerSec*1*2;	//is 16 bit data
	Wfx.nBlockAlign = 1*2;				//number bytes per sample
	Wfx.wBitsPerSample = 16;
	Wfx.cbSize = 0;
	BufSize = 4096;				// buffer size

	switch(m_pDlg->m_SampsPerSec)	//determine soundcard buffersizes
	{							//(need a couple hundred milliseconds worth)
		case 8000:
		case 9600:
		case 11025:
			dwOutBlockSize = BufSize;	//soundcard buffer size
			break;
		case 22050:
			dwOutBlockSize = BufSize*2;	//soundcard buffer size
			break;
		case 44100:
			dwOutBlockSize = BufSize*4;	//soundcard buffer size
			break;
		case 48000:
			dwOutBlockSize = BufSize*4;	//soundcard buffer size
			break;
	}

//allocate some buffers for signals
	if( (m_pOutBuf = new double[sizeof(double)*BufSize]) == NULL )
	{
		AfxMessageBox( _T("Output Memory Error") , MB_ICONSTOP );
		StopMyself();
	}

	SampleLimit = m_pDlg->m_TimeLimit * m_pDlg->m_SampsPerSec;
//instantiate either a soundcard or wavefile object for output.
	if( m_pDlg->m_OutputGroup == SOUNDCARD_OUTPUT )
	{
		m_pSoundDev = new CSound;
		ErrorCode = m_pSoundDev->OutOpen(  &Wfx, dwOutBlockSize, 0, m_pDlg->m_SoundCardSelect);
//		ErrorCode = m_pSoundDev->OutOpen(  &Wfx, dwOutBlockSize, 0,1);
	}
	else
	{
		m_pWaveDev = new CWave;
		ErrorCode = m_pWaveDev->OutOpen( &m_pDlg->m_PathName , &Wfx,
							dwOutBlockSize, SampleLimit,TRUE);
		SamplesSaved = 0;
	}
	if( ErrorCode != NO_ERROR )
	{
		DisplayError( ErrorCode );
		StopMyself();
	}
	CClientDC dc(m_pDlg);			// get a screen device context for the
									// SSTV image generator functions

	m_pGenSrc = new CGenerators;		// use for sin and noise generation
// Initialize the correct generator that is to be used
	switch(m_pDlg->m_SourceGroup)
	{
		case SIN_GEN:
			m_pGenSrc->InitNCO( m_pDlg->m_SampsPerSec, 
								(double)m_pDlg->m_SinFreq,
								m_pDlg->m_SweepRate,
								m_pDlg->m_SweepRange );
			break;
		case SSTV_GEN:
			m_pSSTVSrc = new CSSTVGen;
			if( m_pDlg->m_SSTVMode == 0 )	//if S1 mode
				m_pSSTVSrc->InitSSTV( SSTV_S1_VIS_CODE,m_pDlg->m_SampsPerSec, &dc );
			else						// else is M1 mode
				m_pSSTVSrc->InitSSTV( SSTV_M1_VIS_CODE,m_pDlg->m_SampsPerSec, &dc );
			break;
		case PSK_GEN:
			m_pPSKSrc = new CPSKGen;
			m_pPSKSrc->InitPSK( m_pDlg->m_SampsPerSec, (double)m_pDlg->m_SinFreq, m_pDlg->m_PSKmode|(m_pDlg->m_PskSpeedGrp<<2));
			break;
		case BPSK_400A:
			m_p400BPSKSrc = new C400BPSK;
			m_p400BPSKSrc->InitMod( m_pDlg->m_SampsPerSec, (double)m_pDlg->m_SinFreq, FALSE);
			break;
		case BPSK_400B:
			m_p400BPSKSrc = new C400BPSK;
			m_p400BPSKSrc->InitMod( m_pDlg->m_SampsPerSec, (double)m_pDlg->m_SinFreq, TRUE);
			break;
		case RTTY_GEN:
			m_pRTTYSrc = new CRTTYGen;
			m_pRTTYSrc->InitRTTY( m_pDlg->m_SampsPerSec, (double)m_pDlg->m_SinFreq);
			break;
		case MSK_GEN:
			m_pMSKSrc = new CMSKGen;
			m_pMSKSrc->InitMSK( m_pDlg->m_SampsPerSec, (double)m_pDlg->m_SinFreq);
			break;
		case MFSK_GEN:
			m_pMFSKSrc = new CMFSKGen;
			m_pMFSKSrc->InitMFSK( m_pDlg->m_SampsPerSec, (double)m_pDlg->m_SinFreq);
			break;
	}
//
///////////////   M a i n    P r o c e s s i n g   L o o p   ///////////////	
//
	while( !m_fProcThreadQuit)
	{
//Generate Bufsize data samples
		switch(m_pDlg->m_SourceGroup)
		{
			case SIN_GEN:
				m_pGenSrc->InitNCO( m_pDlg->m_SampsPerSec,
									(double)m_pDlg->m_SinFreq,
									m_pDlg->m_SweepRate,
									m_pDlg->m_SweepRange );
				m_pGenSrc->CalcNCO( m_pOutBuf, BufSize );
				break;
			case SSTV_GEN:
				fFinished = m_pSSTVSrc->CalcSSTV( m_pOutBuf , BufSize);
				break;
			case PSK_GEN:
				m_pPSKSrc->SetPSKFreq(m_pDlg->m_SampsPerSec,
									(double)m_pDlg->m_SinFreq,
									m_pDlg->m_SweepRate,
									m_pDlg->m_SweepRange, 
									m_pDlg->m_PskIdleOnly);
				fFinished = m_pPSKSrc->CalcPSK( m_pOutBuf, BufSize, m_pDlg->m_IMDGain );
				break;
			case BPSK_400A:
				m_p400BPSKSrc->SetFreq(m_pDlg->m_SampsPerSec,
									(double)m_pDlg->m_SinFreq,
									m_pDlg->m_SweepRate,
									m_pDlg->m_SweepRange );
				m_p400BPSKSrc->Calc400BPSK(m_pOutBuf, BufSize);
				break;
			case BPSK_400B:
				m_p400BPSKSrc->SetFreq(m_pDlg->m_SampsPerSec,
									(double)m_pDlg->m_SinFreq,
									m_pDlg->m_SweepRate,
									m_pDlg->m_SweepRange );
				m_p400BPSKSrc->Calc400BPSK(m_pOutBuf, BufSize);
				break;
			case RTTY_GEN:
				m_pRTTYSrc->SetRTTYFreq(m_pDlg->m_SampsPerSec,
									(double)m_pDlg->m_SinFreq,
									m_pDlg->m_SweepRate,
									m_pDlg->m_SweepRange, m_pDlg->m_PskIdleOnly, 
                           m_pDlg->m_Reverse);
				m_pRTTYSrc->CalcRTTY(m_pOutBuf, BufSize);
				break;
			case MFSK_GEN:
				m_pMFSKSrc->SetMFSKFreq(m_pDlg->m_SampsPerSec,
									(double)m_pDlg->m_SinFreq,
									m_pDlg->m_SweepRate,
									m_pDlg->m_SweepRange, m_pDlg->m_PskIdleOnly, 
                           m_pDlg->m_Reverse);
				m_pMFSKSrc->CalcMFSK(m_pOutBuf, BufSize);
				break;
			case MSK_GEN:
				m_pMSKSrc->SetMSKFreq(m_pDlg->m_SampsPerSec,
									(double)m_pDlg->m_SinFreq,
									m_pDlg->m_SweepRate,
									m_pDlg->m_SweepRange );
				m_pMSKSrc->CalcMSK(m_pOutBuf, BufSize);
				break;
		}
		if(fFinished)
		{
			StopMyself();
		}
// Generate BufSize Gaussian noise samples and add to signal with S/N values
		EnterCriticalSection(&vCriticalSection);
		m_pGenSrc->AddBWLimitedNoise(BufSize, m_pOutBuf, m_pDlg->m_SignalAmplitude,
									m_pDlg->m_NoiseAmplitude,m_pDlg->m_BWLimit );

		LeaveCriticalSection(&vCriticalSection);
		m_pDlg->m_RMSOutput = (INT)CalcRMS(m_pOutBuf, BufSize);
//output data to soundcard object or wave file object
		if( m_pDlg->m_OutputGroup == SOUNDCARD_OUTPUT )
		{
			if( (result = m_pSoundDev->OutWrite( m_pOutBuf, BufSize )) != BufSize)
			{
				if( result != 0)
				{
					ErrorCode = m_pSoundDev->GetError();
					DisplayError( ErrorCode );
				}
				m_pSoundDev->OutClose();
				StopMyself();
			}
		}
		else
		{
			if( (result = m_pWaveDev->OutWrite( m_pOutBuf, BufSize )) != BufSize)
			{
				if( result != 0)
				{
					ErrorCode = m_pWaveDev->GetError();
					DisplayError( ErrorCode );
				}
				m_pWaveDev->OutClose();
				StopMyself();
			}
			SamplesSaved += BufSize;
			EnterCriticalSection(&vCriticalSection);
			m_pDlg->m_RunningTime = ((SampleLimit-SamplesSaved)/m_pDlg->m_SampsPerSec);
			LeaveCriticalSection(&vCriticalSection);
		}
	}
// here if main thread has signaled this thread to shut down
	DeleteResources();		//kill all resources
}


///////////////////////////////////////////////////////////////////////////
//   Called to delete any allocated resources in this class
///////////////////////////////////////////////////////////////////////////
void CProcThread::DeleteResources()
{
//  if any object pointer is active, kill it.
	if(	m_pSoundDev)
	{
		m_pSoundDev->OutClose();
		delete m_pSoundDev;
		m_pSoundDev = NULL;
	}
	if(	m_pWaveDev)
	{
		m_pWaveDev->OutClose();
		delete m_pWaveDev;
		m_pWaveDev = NULL;
	}
	if(	m_pGenSrc)
	{
		delete m_pGenSrc;
		m_pGenSrc = NULL;
	}
	if(	m_pPSKSrc)
	{
		delete m_pPSKSrc;
		m_pPSKSrc = NULL;
	}
	if(	m_pMSKSrc)
	{
		delete m_pMSKSrc;
		m_pMSKSrc = NULL;
	}
	if(	m_pMFSKSrc)
	{
		delete m_pMFSKSrc;
		m_pMFSKSrc = NULL;
	}
	if(	m_pRTTYSrc)
	{
		delete m_pRTTYSrc;
		m_pRTTYSrc = NULL;
	}
	if(	m_pSSTVSrc)
	{
		delete m_pSSTVSrc;
		m_pSSTVSrc = NULL;
	}
	if(m_p400BPSKSrc)
	{
		delete m_p400BPSKSrc;
		m_p400BPSKSrc = NULL;
	}
	if(m_pOutBuf)
	{
		delete m_pOutBuf;	
		m_pOutBuf = NULL;
	}
}

/////////////////////////////////////////////////////////////////////
// Simple minded error message display for various sound/wave errors
// Should be put in string table.
void CProcThread::DisplayError( INT ErrorCode )
{
CString Msg;
	switch( ErrorCode )
	{
		case NO_ERRORS:
			Msg = _T("");
			break;
		case MEMORY_ERROR:
			Msg = _T("Memory Allocation error");
			break;
		case WAVIN_ERR_OPEN:
			Msg = _T("Can't open wave file for input");
			break;
		case WAVIN_ERR_NOTWAVE:
			Msg = _T("File is not a RIFF wave type");
			break;
		case WAVIN_ERR_INVALID:
			Msg = _T("Invalid wave file");
			break;
		case WAVIN_ERR_NODATA:
			Msg = _T("No data in file");
			break;
		case WAVIN_ERR_NOTSUPPORTED:
			Msg = _T("Not a supported data type");
			break;
		case WAVIN_ERR_READING:
			Msg = _T("Error reading data from file");
			break;
		case WAVIN_ERR_NOTOPEN:
			Msg = _T("Tried to read and file is not open");
			break;
		case WAVOUT_ERR_OPEN:
			Msg = _T("Can't open wave file for input");
			break;
		case WAVOUT_ERR_WRITING:
			Msg = _T("Error writing to wave file");
			break;
		case WAVOUT_ERR_NOTOPEN:
			Msg = _T("Tried to write and file is not open");
			break;
		case SOUNDIN_ERR_NOTOPEN:
			Msg = _T("Tried to read and soundcard is not open");
			break;
		case SOUNDIN_ERR_OVERFLOW:
			Msg = _T("Input buffers overflowed");
			break;
		case SOUNDIN_ERR_TIMEOUT:
			Msg = _T("Timed out waiting for input buffers");
			break;
		case SOUNDOUT_ERR_NOTOPEN:
			Msg = _T("Tried to write and soundcard is not open");
			break;
		case SOUNDOUT_ERR_UNDERFLOW:
			Msg = _T("Output buffers Underflowed");
			break;
		case SOUNDOUT_ERR_TIMEOUT:
			Msg = _T("Timed out waiting for output buffers");
			break;
//
		case MMSYSERR_BADDEVICEID:
			Msg = _T("Device ID out of range");
			break;
		case MMSYSERR_NOTENABLED:
			Msg = _T("Driver failed enable");
			break;
		case MMSYSERR_ALLOCATED:
			Msg = _T("Device already allocated");
			break;
		case MMSYSERR_INVALHANDLE:
			Msg = _T("Device handle is invalid");
			break;
		case MMSYSERR_NODRIVER:
			Msg = _T("No device driver present");
			break;
		case MMSYSERR_NOMEM:
			Msg = _T("mmsystem memory allocation error");
			break;
		case MMSYSERR_NOTSUPPORTED:
			Msg = _T("Function isn't supported");
			break;
		case MMSYSERR_BADERRNUM:
			Msg = _T("Error value out of range");
			break;
		case MMSYSERR_INVALFLAG: 
			Msg = _T("Invalid flag passed");
			break;
		case MMSYSERR_INVALPARAM: 
			Msg = _T("Invalid parameter passed");
			break;
		case MMSYSERR_HANDLEBUSY:
			Msg = _T("Handle being used");
			break;
		default:			// unknown error code
			Msg = _T("??? Maybe Input being used at\n a different sample rate");
			break;
	}
	if( Msg != _T("") )
		AfxMessageBox( Msg );
}

///////////////////////////////////////////////////////////////////
//calculates RMS value of buf[] samples over RMS_SIZE*numsamps
//  called once for each buffer to be measured
///////////////////////////////////////////////////////////////////
double CProcThread::CalcRMS(double* buf, INT bufsize)
{
#define RMS_SIZE 10
static double sum = 0.0;
static int cnt = 0;
static double rms = 0.0;
	for( INT i = 0; i<bufsize; i++)
		sum = sum + ( buf[i]*buf[i] );
	if( ++cnt >= RMS_SIZE)
	{
		rms = sqrt(sum/(cnt*bufsize));
		sum = 0.0;
		cnt = 0;
	}
	return rms;
}